/**********************************************************************
*  File:    HOST.C                                                    *
*  Purpose: Comtrol Hostess i Programmers' Toolkit, Main Program and  *
*           Service functions.                                        *
*  Date:    January 2, 1990                                           *
*  Author:  Howard D. Straus                                          *
*  Release: 1.00 - Original release                                   *
***********************************************************************
*                         COPYRIGHT NOTICE                            *
*=====================================================================*
*   Copyright 1990, 1991 by Howard D. Straus and Comtrol Corporation. *
*   All rights reserved.  Use or reproduction by any means of any     *
*   portion of this code contrary to the developers license agreement *
*   is expressly prohibited.                                          *
***********************************************************************/

#include "compiler.h"

#include <dos.h>
#ifdef  TURBO_C
#include <graphics.h>
#else
#include <graph.h>
struct  rccoord temp;
#endif
#include <conio.h>
#include <stdio.h>
#include <time.h>

#include "..\hostess\hostess.h"

#include "host.h"
#include "hostglb.h"
#include "hostfns.h"

char    entry_screen[4000];             /* save screen on entry */
int     orig_x;                 /* save text position on entry */
int     orig_y;

char    parity_chr[] = "N?OE????";

char    *stop_bit_str[] = { "?", "?", "1", "1.5", "2" };

char    *error_msg[] = { "No error message.",
                "Rx characters must have 7 or 8 data bits.",
                "Parity required characters with 7 data bits.",
                "Tx characters must have 7 or 8 data bits.",
                "Attempt to open an unconfigured port.",
                "Attempt to write to a closed line.",
                "Attempt to open an active line.",
                "Unknown host command received."
};


/***********************************************************************
*  Name:        main                                                   *
*  Purpose:     main monitor for the Comtrol Hostess i Communications  *
*       Coprocessor Board C Programmer's Toolkit system                *
*  Call:        invoked from the command line, with command line pa-   *
*       rameters.                                                      *
*  Parameters:  on command line                                        *
*  Returns:     none.                                                  *
*  Method:      after initialization of screen, memory and Hostess, an *
*       endless loop is performed, in which each coprocessor board or  *
*       user keyboard function is continually checked.                 *
*  Notes:                                                              *
************************************************************************/

main(argc, argv)
        int     argc;
        char    **argv;
{       int     skip_count = 0;

        handle_command_line(argc, argv);
        initialize();
        for (;;)
        {   if (++skip_count > 300)     /* cleanup for double interrupts */
                {       skip_count = 0;                 /* missed interrupts, etc */
                        if ((argc = hcb_buffer_count()) != 0) /* anything pending */
                                receive_hcb_message(); /* pick it up and process it */
                }
                if (hcb_interrupt)          /* if an interrupt is pending */
                {   if (hcb_buffer_count()) /* check for message from Hostess */
                                receive_hcb_message(); /* get one if present */
                        else
                        {   DISABLE();          /* all done, disable interrupts */
                                hcb_interrupt = 0;  /* clear interrupts pending count */
                                ENABLE();           /* restore interrupt system */
                        }
                }
                if (!hq_qem(&output_qcb))
                        if (!host_buffer_count())
                                send_next_buffer();
                check_deferred_items();
                if (kbhit())
                        handle_key(get_key());
                if ((kp != keyboard_buffer) && (time(NULL) != last_second))
                        send_char_to_hcb(0);
        }
}


/***********************************************************************
*  Name:        check_deferred_items                                   *
*  Purpose:     to monitor transmissions which were deferred, and      *
*       send them as soon as resources become available.               *
*  Call:        check_deferred_items();                                *
*  Parameters:  none.                                                  *
*  Returns:     none.                                                  *
*  Method:      if a deferred item flag is set, an attempt is made to  *
*       resend the message.                                            *
*  Notes:                                                              *
************************************************************************/

void
check_deferred_items()
{
        if (set_time_flag)
                set_time();
        if (modem_control_flag)
                send_modem_control(modem_control_word);
}


/***********************************************************************
*  Name:        get_key                                                *
*  Purpose:     fetches and interprets the latest keystroke            *
*  Call:        c = get_key();                                         *
*  Parameters:  none.                                                  *
*  Returns:                                                            *
*       UINT c;  the key which was struck.  If the key was a function  *
*           key (multi-byte), 0x100 is added to the character.         *
*  Method:      if the first character returned is 0, and there is     *
*       still a key character pending, the character is a function     *
*       key, and has the function code added to it.                    *
*  Notes:                                                              *
************************************************************************/

UINT
get_key()
{   UINT    c;

        if (((c = getch()) == 0) && kbhit())
           c = FUNCT | getch();
        return(c);
}


/***********************************************************************
*  Name:        handle_command_line                                    *
*  Purpose:     interpret and decode the various flags in the command  *
*       line, placing variables into their assigned places.            *
*  Call:        handle_command_line(argc, argv);                       *
*  Parameters:                                                         *
*       int argc;  command line argument count (from main)             *
*       char *argv[];  array of pointers to the argument strings       *
*  Returns:     none.                                                  *
*  Method:      interprets the command line arguments, one by one,     *
*       from right to left.  Decodes the string, hex or integer ar-    *
*       guments, and places them in the correct places, overriding     *
*       the defaults.                                                  *
*  Notes:       each argument flag must have a '-' prepended to it.    *
*       In case of argument error, the command flag menu will be       *
*       printed on the screen, and the program exits.                  *
************************************************************************/

void
handle_command_line(argc, argv)
        int     argc;
        char    *argv[];
{   char    c;

        while (--argc)
        {   if (*argv[argc]++ != '-')   /* first character must be '-' */
                {   show_help();
                        exit(-1);
                }

                switch (toupper(*argv[argc]++)) {

                case 'B':                   /* refer to board number vs address */
                        board_no = atoi(argv[argc]) - 1; /* convert to integer */
                        if (board_no >= MAX_BOARD_NO)
                        {   printf("tk: illegal board number (%d)\n", board_no + 1);
                                show_help();
                                exit(-1);
                        }
                        base_address = default_base[board_no];
                        irq_no = default_irq[board_no];
                        base_port = io_port[board_no];
                        break;

                case 'C':                   /* color or monochrome */
                        if (toupper(*argv[argc]) == 'C') /* if color */
                                iscolor = TRUE;     /* default */
                        else
                                iscolor = FALSE;    /* monochrome adaptor */
                        break;

                case 'D':                   /* data bits */
                        data_bits = atoi(argv[argc]); /* data bits to integer */
                        if ((data_bits < 7) || (data_bits > 8))
                        {   printf("tk: illegal number of data bits (%d)\n", data_bits);
                                show_help();
                                exit(-1);
                        }
                        break;

                case 'F':                   /* message file name */
                        strcpy(message_file_name, argv[argc]); /* copy file name */
                        break;

                case 'I':                   /* IRQ number */
                        irq_no = atoi(argv[argc]); /* convert to integer */
                        break;

                case 'L':                   /* Hostess comm port select */
                        hcb_port = atoi(argv[argc]) - 1;
                        if ((hcb_port < 0) || (hcb_port > 15))
                        {   show_help();
                                exit(-1);
                        }
                        break;

                case 'M':                   /* shared memory base address */
                        base_address = atoh(argv[argc]); /* convert to long */
                        if ((base_address > 0xF00000L) ||
                                 ((base_address & 0x1FFFFL) != 0))/* illegal address */
                        {   printf("tk: illegal base address: %s\n", *argv[argc]);
                                show_help();
                                exit(-1);
                        }
                        break;

                case 'O':                   /* base port select */
                        base_port = atoh(argv[argc]); /* convert hex to integer */
                        if ((base_port < 0x200) || (base_port > 0x3FF))
                        {   printf("tk: illegal base port number: %s\n", *argv[argc]);
                                show_help();
                                exit(-1);
                        }
                        break;

                case 'P':                   /* parity select */
                        switch (toupper(*argv[argc])) {  /* what kind of parity */

                        case 'E':               /* even parity */
                                parity = EVEN;
                                break;

                        case 'O':               /* odd parity */
                                parity = ODD;
                                break;

                        case 'N':               /* no parity */
                                parity = NONE;
                                break;

                        default:
                                printf("tk: illegal parity selector (%c)\n", *argv[argc]);
                                show_help();
                                exit(-1);
                                break;
                        }
                        break;

                case 'S':                   /* port speed */
                        port_speed = atoi(argv[argc]); /* convert to integer */
                        if ((port_speed > 9600) || (port_speed < 50))
                        {   printf("tk: illegal port speed %d\n", port_speed);
                                show_help();
                                exit(-1);
                        }
                        break;

                case 'X':                   /* executable load file name */
                        strcpy(hcb_load_file_name, argv[argc]);
                        break;

                case 'Z':                   /* stop bits */
                        if ((c = *argv[argc]++) == '2')
                                stop_bits = 4;      /* two stop bits */
                        if ((c == '1') && (*argv[argc] == '.'))
                                stop_bits = 3;      /* one and a half stop bits */
                        else
                                stop_bits = 2;      /* one stop bit (default) */
                        break;

                default:
                        printf("tk: unknown command line option (%s)\n",
                                        *(--argv[argc]));
                        show_help();
                        exit(-1);
                }
        }
        if ((parity == NONE) && (data_bits < 8))
        {   printf("tk: data bits plus parity must add up to 8 or more bits.\n");
                exit(-1);
        }
}


/***********************************************************************
*  Name:        handle_key                                             *
*  Purpose:     handles keystrokes entered during program execution    *
*  Call:        handle_key(c);                                         *
*  Parameters:                                                         *
*       int c;  the key which was struck.  If key is > 0x100, it is    *
*           a function key.                                            *
*  Returns:     none.                                                  *
*  Method:      one large switch statement executes the desired func-  *
*       tion.                                                          *
*  Notes:       for mnemonic purposes, more than one keystroke can     *
*       execute the same function.                                     *
************************************************************************/

void
handle_key(c)
        UINT            c;      /* the key which was struck */
{       int             *sp;
        int     far     *lp;
        UINT            i;

        if (saved_ptr != NULL)  /* anything in auxiliary window */
        {       restore_screen();
                show_output("");
                in_aux = 0;                     /* clear aux window contents */
                return;
        }

        switch (c) {

        case ALT_D:         /* toggle DTR */
        case F6:
                modem_control_word ^= DTR;
                send_modem_control(modem_control_word);
                break;

        case ALT_F:         /* send a text file to a port */
        case F4:
                send_file();
                break;

        case ALT_H:         /* Help */
        case F1:
                show_help_screen(level); /* show help screen */
                break;

        case ALT_L:         /* load Hostess */
        case F3:
                initialize_hcb(); /* initialize Hostess operation */
                break;

        case ALT_M:         /* memory dump */
                dump_memory();
                break;

        case ALT_N:                     /* show line table */
                show_line_table();
                break;

        case ALT_O:         /* enter on-board debugger */
        case F8:
                enter_debugger();
                break;

        case ALT_P:         /* toggle loopback mode */
        case F7:
                loopback_mode ^= TRUE;
                configure_port();
                open_line(3);   /* open line for input and output */
                break;

        case ALT_R:         /* reset Hostess */
        case F10:
                reset_hcb();    /* reset the Hostess board */
                break;

        case ALT_S:         /* toggle RTS */
        case F5:
                modem_control_word ^= RTS;
                send_modem_control(modem_control_word);
                break;

        case ALT_T:         /* set time */
        case F9:
                while (host_buffer_count())
                        ;           /* wait for the host buffer to free up */
                set_time();     /* call the set time utility */
                break;

        case ALT_V:         /* show device control table */
        case F2:
                show_device_table();
                break;

        case ALT_X:         /* quit, restore original screen */
        case ESC:
#ifdef  TURBO_C
                puttext(1, 1, 80, 25, entry_screen);
#else
                sp = (int *)entry_screen;
                lp = (int far *)videomemory_addr;
                for (i = 0; i < 2000; i++)
                        *lp++ = *sp++;
#endif
                TEXTPOSITION(orig_x, orig_y);
                exit(0);
                break;

        default:
                if (c < 0x7F)       /* character to send to Hostess port */
                        send_char_to_hcb(c & 0x7F);  /* send it out */
                else                            /* illegal character */
                        putch('\007');  /* sound alarm */
                break;
        }
}


/************************************************************************
*  Name:        initialize                                              *
*  Purpose:     sets up the operation of the system based on the setup  *
*       parameters and defaults.  Establishes communications with the   *
*       Hostess board.                                                  *
*  Call:        initialize();                                           *
*  Parameters:  none.                                                   *
*  Returns:     none.                                                   *
*  Method:      uses the default and command line parameters to con-    *
*       figure the system, establishes communications with the Hostess, *
*       uploads the selected executable program, starts it running,     *
*       and insures that it is running properly.                        *
*  Notes:                                                               *
*************************************************************************/

void
initialize()
{   int     temp_word;
        UINT            i;
        UINT                    mask;
        int             far             *lp;
        int                             *sp;

#ifdef  TURBO_C
        gettext(1, 1, 80, 25, entry_screen);
#else
        sp = (int *)entry_screen;
        videomemory_addr = (iscolor ? MK_FP(0xB800, 0) : MK_FP(0xB000, 0));
        lp = (int far *)videomemory_addr;
        for (i = 0; i < 2000; i++)
                *sp++ = *lp++;
#endif
        orig_y = GETTEXTROW();
        orig_x = GETTEXTCOL();

        hcb_interrupt = 0;            /* initialize pending interrupts */
        if (irq_no < 8)
        {   save_vector = (void far *)GET_VECTOR(irq_no + 8);
                SET_VECTOR(irq_no + 8, hcb_interrupt_service);
                mask = INP(INTMASK_LO);         /* get current interrupt mask */
                mask &= ~(1 << irq_no);         /* mask off bit to enable interrupt */
                OUTP(INTMASK_LO, mask);         /* put mask in control register */
        }
        else
        {   save_vector = (void far *)GET_VECTOR(irq_no + 0x68);
                SET_VECTOR(irq_no + 0x68, hcb_interrupt_service);
                mask = INP(INTMASK_HI);         /* get current interrupt mask */
                mask &= ~(1 << (irq_no - 8)); /* mask off bit to enable interrupt */
                OUTP(INTMASK_HI, mask);         /* put mask in control register */
        }
        modem_control_word = 0;         /* output modem control word */
        init_windows();                 /* set up the screen */
        show_modem_controls(0);         /* set up the modem control area */
        show_time(0L);                  /* set up the time area */
        change_to_window(1);            /* diagnostic area */
        hq_ini();                       /* output queue control block */
        hq_qci(&output_qcb);            /* free queue control block */
        initialize_hcb();               /* begin Hostess operations */
}


/************************************************************************
*  Name:        initialize_hcb                                          *
*  Purpose:     loads and starts up the Hostess board, based on the     *
*       stored parameters.                                                                                              *
*  Call:        initialize_hcb();                                       *
*  Parameters:  none.                                                   *
*  Returns:     none.                                                   *
*  Method:      uses the parameters set up by default and command line  *
*       to set up the communications on the Hostess i, and begin        *
*       operations with the board.                                      *
*  Notes:                                                               *
*************************************************************************/

void
initialize_hcb()
{   ULONG   scc_map_address;        /* port availability map address */
        UINT    scc_map;                /* port availability map */
        UINT    er;                     /* read error from DP-RAM */

        reset_hcb();                    /* reset the Hostess card */
        scc_map_address = base_address + SCC_MAP_OFFSET;
        read_from_hcb((char far *)&scc_map, scc_map_address, 1);
                                                                        /* read port availability map */
        if (!((1 << hcb_port) & scc_map))
        {   show_output("<< Selected Hostess port not available. >>\n");
                exit(-1);
        }
        if (load_hcb(hcb_load_file_name)) /* load with executable program */
        {       show_output("<< Failure loading executable file. >>\n");
                exit(-1);
        }
        if (begin_hcb_execution())      /* start execution */
        {   show_output("<< Hostess failed to begin execution. >>\n");
/*      exit(-1);  */
        }

        mailbox_addr = base_address + HIGH_PAGE + MAILBOX_OFFSET;
                                                                        /* initialize mailbox address */
        read_from_hcb((char far *)&mailbox, mailbox_addr, 8);
                                                                        /* initial read of mailbox */
        host_buffer_addr = base_address + norm(mailbox.mb_homg);
        hcb_buffer_addr = base_address + norm(mailbox.mb_ulmg);
        configure_port();                               /* configure the I/O port */
        open_line(3);                                   /* open port for input and output */
        set_time();                     /* set the Hostess time of day */
}


/***********************************************************************
*  Name:        process_hcb_message                                    *
*  Purpose:     process the message sent to the host by the Hostess    *
*       control program.                                               *
*  Call:        process_hcb_message(mg, ct);                           *
*  Parameters:                                                         *
*       MG_ *mg;  pointer to the message structure containing the      *
*               data from the Hostess board.                           *
*       int ct;  number of bytes of data in the message data area.     *
*  Returns:     none.                                                  *
*  Method:      determines whether data or status is being received,   *
*       then displays the received data in the appropriate window on   *
*       the formatted screen.                                          *
*  Notes:       receives the line and device control blocks for later  *
*       display.                                                       *
************************************************************************/

void
process_hcb_message(mg, ct)
        MG_     *mg;
        int     ct;
{   RB_     *rb;                    /* received data buffer structure */
        RD_             *rd;                                    /* raw data buffer */
        LS_     *ls;                    /* line status message structure */
        DV_     *dv;                    /* device control block structure */
        LC_             *lc;                                    /* line control block structure */
        int             l;                                              /* work variable */

        switch (mg->mg_type) {

        case DATA_BUFFER:               /* data received from port */
                rb = (RB_ *)mg;                         /* point to message */
                show_input(rb->rb_data);    /* put received data in input window */
                l = strlen(rb->rb_data);        /* find string length */
                if (rb->rb_data[l - 1] == '\r') /* if last char is CR */
                        show_input("\n");               /* show a LF */
                if (rb->rb_data[l - 1] == '\n') /* if last char is LF */
                        show_input("\r");               /* show a CR */
                if (strlen(log_file_name) != 0) /* if log file desired */
                        fputs(rb->rb_data, log_file); /* log incoming data */
                break;

        case STATUS_REPORT:             /* status message */
                ls = (LS_ *)mg->mg_data;    /* point to line status message */
                dv = (DV_ *)&ls->ls_dctb;   /* point to imbedded device ctl table */
                lc = (LC_ *)&ls->ls_lctb;       /* point to imbedded line ctl table */
                show_time(ls->ls_time);     /* display the latest Hostess time */
                show_modem_controls(((dv->dv_rts | dv->dv_dtr) << 8) | dv->dv_lrr0);
                memcpy(&lct, lc, sizeof(LC_)); /* copy tables to local buffers */
                memcpy(&dct, dv, sizeof(DV_));
                if (in_aux == LINE_TABLE)       /* if line table in aux window */
                        show_line_table();              /* interpret it */
                if (in_aux == DEVICE_TABLE)     /* if device table is up */
                        show_device_table();    /* interpret it */
                break;

        case ERROR:                     /* error message */
                rd = (RD_ *)mg->mg_data;    /* point to message data */
                mg->mg_data[ct] = 0;        /* insure null delimiter */
                show_input("<< ");          /* differentiate from data */
                show_input(error_msg[rd->rd_data[0]]); /* interpret received error */
                show_input(" >>\n\r");        /* message into the input window */
                break;

        default:
                break;
        }
}


/***********************************************************************
*  Name:        receive_hcb_message                                    *
*  Purpose:     fetches a message from the Hostess dual-ported memory  *
*       then processes the contents.                                   *
*  Call:        receive_hcb_message();                                 *
*  Parameters:  none.                                                  *
*  Returns:     none.                                                  *
*  Method:      reads up a message from the Hostess dual-ported RAM    *
*       host communications area to a local allocated message buffer,  *
*       then passes it to the process_hcb_message function.  Re-       *
*       leases the allocated memory on exit.                           *
*  Notes:                                                              *
************************************************************************/

void
receive_hcb_message()
{   MG_     *mg;
        int     byte_count;
        int     word_count;
        int     zero;

        zero = 0;                   /* need a word of zero */
        /*
                first read in the data count (in bytes) from the Hostess
                mailbox host semaphore
        */
        if (read_from_hcb((char far *)&byte_count, mailbox_addr + 2, 1))
        {   show_output("<< Error reading count from Hostess. >>\n");
                return;
        }
        if (byte_count == 0)        /* no data in host comm area */
                return;
        /*
                now read to a local message buffer the contents of the Hostess
                data communications buffer
        */
        word_count = (byte_count + 10) >> 1; /* read trailing zeros */
        mg = (MG_ *)malloc(byte_count + 10); /* get a buffer into which to read */
        if (read_from_hcb((char far *)mg, hcb_buffer_addr, word_count))
        {   show_output("<< Error reading Hostess buffer. >>\n");
                return;
        }
        /*
                finally, reset the Hostess data semaphore in the Hostess's mailbox
                to indicate that the host has picked up the data, and the
                buffer is now empty
        */
        if (write_to_hcb(mailbox_addr + 2, (char far *)&zero, 1))
        {   show_output("<< Error clearing Hostess count. >>\n");
                return;
        }
        process_hcb_message(mg, byte_count); /* process the message */
        free(mg);                       /* free the allocated memory */
        return;
}


/***********************************************************************
*  Name:        send_next_buffer                                       *
*  Purpose:     sends the next queued buffer to the Hostess as a       *
*       message                                                        *
*  Call:        send_next_buffer();                                    *
*  Parameters:  none.                                                  *
*  Returns:     none.                                                  *
*  Method:      gets the next buffer from the host output queue, and   *
*       writes the data contents to the host communications (write)    *
*       area in the Hostess dual-ported RAM.  It then sets the message *
*       data count in the mailbox to the length of the message.        *
*  Notes:                                                              *
************************************************************************/

void
send_next_buffer()
{   BH_     *bh;                        /* buffer pointer */
        int     word_count;                 /* words to send to Hostess */

        if ((bh = hq_deq(&output_qcb)) == NULL) /* dequeue next buffer */
                return;                         /* none on queue */

        word_count = (bh->bh_txct+1) >> 1;  /* data words to transfer */
        /*
                first write the data from the buffer to the Hostess's Host
                Communications Buffer.
        */
        if (write_to_hcb(host_buffer_addr, (char far *)bh->bh_text, word_count))
        {   show_output("<< Error writing buffer to Hostess >>\n");
                return;
        }
        /*
                now set the semaphore in the Hostess mailbox to the message
                data count (in bytes).  This is the signal to the Hostess that
                there is new data in the buffer.
        */
        if (write_to_hcb(mailbox_addr, (char far *)&bh->bh_txct, 1))
        {   show_output("<< Error writing count to Hostess >>\n");
                return;
        }
        OUTP(base_port + HCB_START_PORT_OFFSET, 0); /* interrupt Hostess */
        hq_fre(bh);                                     /* free the buffer */
}


/********************************************************************
*  Name:        show_help                                           *
*  Purpose:     displays the command line parameters on the screen  *
*  Call:        show_help();                                        *
*  Parameters:  none.                                               *
*  Returns:     none.                                               *
*  Method:      simply writes a string to the screen.               *
*  Notes:                                                           *
*********************************************************************/

void
show_help()
{
        printf("usage: tk -BDFILMOPSXZ\n\
        -Bn       Hostess board (1-7) [%d]\n\
        -Cc       Color / monochrome adaptor (C/M) [%c]\n\
        -Dn       number of data bits in character (7-8) [%d]\n\
        -Ff       message file [%s]\n\
        -In       IRQ (3,4,5,10,11,15) [%d]\n\
        -Ln       port number (1-16) [%d]\n\
        -Maddr    shared memory base address [%lX]\n\
        -Oadd     board base I/O port (hex) [%X]\n\
        -Pp       port parity (E/O/N) [%c]\n\
        -Sbps     port speed (50-9600) [%d]\n\
        -Xfn      executable file name [%s]\n\
        -Zn       stop bits (1, 1.5, 2) [%s]\n",
        board_no + 1, (iscolor ? 'C' : 'M'), data_bits, message_file_name,
        irq_no, hcb_port + 1, base_address, base_port, parity_chr[parity],
        (port_speed == -1 ? 9600 : port_speed), hcb_load_file_name,
        stop_bit_str[stop_bits]);
}


/***********************************************************************
*  Name:        hcb_interrupt_service                                  *
*  Purpose:     service the interrupt from the Hostess board           *
*  Call:        called from IRQ vector                                 *
*  Parameters:  none.                                                  *
*  Returns:     none.                                                  *
*  Method:      sets the interrupt received flag, then gives the 8259  *
*       PIC an End-of-Interrupt indicator.  If the IRQ is above 8,     *
*       both PICs must be given EOIs, as the interrupts are cascaded.  *
*  Notes:                                                              *
************************************************************************/

void interrupt
hcb_interrupt_service()
{
        ++hcb_interrupt;          /* bump unserviced interrupts count */
        if (irq_no >= 8)            /* if IRQ is on the high controller */
                OUTP(INTCTL_HI, EOI);   /* give high interrupt controller an EOI */
        OUTP(INTCTL_LO, EOI);       /* give low interrupt controller an EOI */
}


                                  /* ***** END OF FILE ***** */

