/**********************************************************************
*  File Name: HOSTESS.H                                               *
*  Purpose:   COMTROL Hostess i Programmer's Toolkit Structures,      *
*       Constants and Definitions                                     *
*  Date:      September 19, 1990                                      *
*  Author:    Howard D. Straus                                        *
*  Release:   1.00 - Original release                                 *
***********************************************************************
*                         COPYRIGHT NOTICE                            *
*=====================================================================*
*   Copyright 1990, 1991 by Howard D. Straus and Comtrol Corporation. *
*   All rights reserved.  Use or reproduction by any means of any     *
*   portion of this code contrary to the developers license agreement *
*   is expressly prohibited.                                          *
***********************************************************************/

#include    <time.h>

#ifndef NULL
#define NULL            0
#endif

#define TRUE            1
#define FALSE           0

#define OK              0
#define ER              1

#define FIRST_FREE_MEMORY 0xC00         /* first unused memory location */

#define TOOLKIT_RUNNING_VALUE 0xBABE    /* unique flag value */

#define BIT0            (1 << 0)        /* bit masks */
#define BIT1            (1 << 1)
#define BIT2            (1 << 2)
#define BIT3            (1 << 3)
#define BIT4            (1 << 4)
#define BIT5            (1 << 5)
#define BIT6            (1 << 6)
#define BIT7            (1 << 7)
#define BIT8            (1 << 8)
#define BIT9            (1 << 9)
#define BIT10           (1 << 10)
#define BIT11           (1 << 11)
#define BIT12           (1 << 12)
#define BIT13           (1 << 13)
#define BIT14           (1 << 14)
#define BIT15           (1 << 15)

/*
        Read register 0 status bits
*/
#define RXA             BIT0
#define TBE             BIT2
#define DCD             BIT3
#define CTS             BIT5
#define BRK_ABORT       BIT7

/*
        Read register 1 status bits
*/
#define PARITY_ERROR    BIT4
#define RX_OVERRUN      BIT5
#define FRAMING_ERROR   BIT6

/*
        Write register 3
*/
#define RXE                             BIT0                    /* Rx enable bit */

/*
        Write register 5
*/
#define TXE                             BIT3                    /* Tx enable bit */

/*
        Write register 14
*/
#define LOOPBACK                BIT4                    /* enable local loopback */

#define BS              1               /* buffer shortage error */

#define N_PORTS         16              /* Hostess i has 16 ports */

#define HIGH_PAGE_SEG   0x1000          /* segment register for high page */
#define MAILBOX         0xFFF0          /* offset of mailbox in high page */
#define HIGH_BOUND      0xFFFF          /* upper bound of available memory */
#define HOST_MSG_SIZE   300             /* maximum data transfer size */

#define RUN             0               /* Tx or Rx running */
#define STOP            1               /* Tx or Rx stopped - flow ctl */

#define DEBUGGER_INT    0x21            /* jump into the ROM debugger */
#define CLEAR_INTERRUPT 0               /* clears host interrupt */

#define PORT_INT        0x80            /* port interrupts start at 0x80 */

#define TIMER_NO        1               /* use timer 1 */

#define HOST_INTERRUPT_PORT 0xEF60      /* port to write for host interrupt */
#define SET_INTERRUPT       0x08        /* sets host interrupt */

#define TIMER_INTR_BASE     52          /* timer interrupt base */
#define TIMER_PORT_BASE     0x9074      /* timer port base */
#define TIMER_PORT_INCR     1           /* timer port spacing */

#define TIME_CONST          31250       /* yields 25 interrupts / sec. */
#define TICKS_PER_SECOND    25          /* timer initialized to 0.04 sec. */

#define PORT0           0xE1F4          /* 8530 base address, port 0 */
#define PORT1           0xE1F0          /* 8530 base address, port 1 */
#define PORT2           0xE3F4          /* 8530 base address, port 2 */
#define PORT3           0xE3F0          /* 8530 base address, port 3 */
#define PORT4           0xE5F4          /* 8530 base address, port 4 */
#define PORT5           0xE5F0          /* 8530 base address, port 5 */
#define PORT6           0xE7F4          /* 8530 base address, port 6 */
#define PORT7           0xE7F0          /* 8530 base address, port 7 */
#define PORT8           0xE9F4          /* 8530 base address, port 8 */
#define PORT9           0xE9F0          /* 8530 base address, port 9 */
#define PORT10          0xEBF4          /* 8530 base address, port 10 */
#define PORT11          0xEBF0          /* 8530 base address, port 11 */
#define PORT12          0xEDF4          /* 8530 base address, port 12 */
#define PORT13          0xEDF0          /* 8530 base address, port 13 */
#define PORT14          0xEFF4          /* 8530 base address, port 14 */
#define PORT15          0xEFF0          /* 8530 base address, port 15 */

#define INT_CTL         0x9070          /* Hostess i 8259 PIC port */

#define TIMER_INTR      (TIMER_INTR_BASE + TIMER_NO)
#define TIMER_PORT      (TIMER_PORT_BASE + TIMER_NO * TIMER_PORT_INCR)
#define TIMER_CTL       (TIMER_PORT_BASE + 3 * TIMER_PORT_INCR)
#define ACTIVATE_TIMER  (0x34 + (TIMER_NO << 6))

#define HOST_INT_PORT   0xFF67          /* host interrupt port */
#define HOST_INT_SET    0x4000          /* set host interupt */
#define HOST_INT_CLR    0               /* clear host interrupt */

#define EXT_INT         48              /* host PC interrupt */

#define INT_INIT_CMD    0xFF52          /* 8259 PIC initialization cmd word */
#define EOI             0x20            /* 8259 PIC end-of-interrupt */

#define CLOCK_RATE      1536000L                /* clock rate for BR generator */
                                                                                /* precomputes (rate * 10 / 32) */
                                                                                /* = (4915200 * 10 / 32) */

#define BUF_SIZE        200             /* message buffer text area size */

/*
        host to Hostess i message types
*/
#define ENTER_DEBUGGER          1
#define SET_TIME                2
#define CONFIG_PORT             3
#define CONFIG_LINE             4
#define OPEN_LINE               5
#define CLOSE_LINE              6
#define WRITE_DATA              7
#define REQ_STATISTICS          8
#define REQ_STATUS              9
#define SEND_BREAK              10
#define MODEM_CONTROL           11
#define SET_DIAGNOSTIC_LEVEL    39      /* set the diagnostic message level */

/*
        Hostess i to host message types
*/
#define DATA_BUFFER             50      /* input characters from comm line */
#define STATUS_REPORT           51      /* line status report per request */
#define STATISTICS_REPORT       52      /* line statistics report, per req. */
#define LINE_ERRORS             59      /* line error occurred */
#define ALARM                   60      /* emergency condition occurred */
#define ERROR                   61      /* host message was in error */
#define DEBUG_MESSAGE           99      /* debug message to display */

/*
        line error messages
*/
#define LE_LINE_BREAK           10      /* break msec exceeded maximum */
#define LE_DCD_LOST             20      /* lost carrier */
#define LE_DCD_ACQUIRED         21      /* acquired carrier */
#define LE_FAILURE              23      /* line failure */
#define LE_FRAMING              30      /* framing error on line */
#define LE_PARITY               31      /* parity error on line */
#define LE_RX_OVR               32      /* receive overrun on line */

/*
        line control module  line states
*/
#define CLOSED           0       /* line is not open */
#define IDLE             1       /* line is idle */
#define ACTIVE           2       /* line is active */
#define IN_BREAK         3       /* line is receiving break condition */
#define LINE_FAIL        4       /* line failure */

/*
        ASCII special characters
*/
#define SOH             1
#define STX             2
#define ETX             3
#define EOT             4
#define CR              '\r'
#define LF              '\n'

/*
        host command consistency errors
*/
#define ER_RX_CHAR_SIZE         1
#define ER_PARITY_REQ           2
#define ER_TX_CHAR_SIZE         3
#define ER_PORT_UNCONFIGURED    4
#define ER_WRITE_TO_CLOSED_LINE 5
#define ER_OPEN_ACTIVE_LINE     6
#define ER_UNKNOWN              7


/*
        alarm conditions
*/
#define AL_NO_BUFFERS           10      /* no buffers available */


/*
        message error byte flags (same as in EXEC)
*/
#define ME_BRK_EOM      BIT8    /* excessive line breaks caused EOM */
#define ME_CTS_CHG      BIT9    /* CTS changed during message */
#define ME_DCD_CHG      BIT10   /* DCD changed during message */
#define ME_FRAMING      BIT11   /* framing error(s) occurred during message */
#define ME_OVERRUN      BIT12   /* receive overrun occurred during message */
#define ME_PARITY       BIT13   /* parity error(s) occurred during message */
#define ME_TIMEOUT      BIT14   /* cessation of reversals caused EOM */


/*
        line control transmit and receive event types
*/
#define NE              0       /* no event */
#define TO              1       /* timeout */
#define LE              2       /* line error */
#define TD              3       /* transmit done */
#define DC              4       /* carrier detect change */
#define CT              5       /* CTS change detect */
#define CL              6       /* host closed line */
#define RE              7       /* device receive end */
#define OP              8       /* host opened line */
#define ST              9       /* start transmission of a message */
#define BK             10       /* break started */
#define BE             11       /* break ended */


/*
        buffer thresholds for buffer level sensitive operations
*/
#define STOP_HOST_RX_THRESH 10  /* stop receiving from host below this */

typedef struct host_message {
        UTINY   mg_type;                /* command/status type */
        UTINY   mg_line;                /* line number concerned */
        UTINY   mg_data[HOST_MSG_SIZE]; /* host message */
} MG_;

#define MG_HDR  (2 * sizeof(UTINY))

typedef struct config_device {  /* device control block configuration */
        UINT          cd_spd;   /* line speed, bits/sec * 10 */
        UTINY         cd_bits;  /* Bits / character */
        UTINY         cd_par;   /* parity: none (0), odd (1), even (3) */
        UTINY         cd_stpb;  /* stop bits: 1 (4), 1.5 (8), 2 (12) */
        UTINY         cd_loop;  /* make internal loop (0x10), no loop (0) */
} CD_;

typedef struct mailbox {
        UINT                   mb_hoct;  /* host buffer data count/semaphore */
        UINT                   mb_ulct;  /* Hostess buffer data count/semaphore */
        MG_     far           *mb_homg;  /* host message buffer addr */
        MG_     far           *mb_ulmg;  /* Hostess message buffer addr */
        struct device_ctl far *mb_dvad;  /* first device control block addr */
} MB_;

typedef struct que_buf {
        struct  que_buf far    *bf_next;    /* forward thread */
        UINT                    bf_txct;    /* buffer text char count */
        UTINY                   bf_type;    /* buffer type */
        UTINY                   bf_lnno;    /* line number */
        time_t                  bf_tmst;    /* timestamp */
        UINT                    bf_stat;    /* indicator flags */
        UINT                    bf_errs;    /* error flags */
        UTINY                   bf_text[BUF_SIZE];
} BF_;

#define BF_HDR (sizeof(int far *)+3*sizeof(int)+sizeof(time_t)+2*sizeof(UTINY))

typedef struct  qcb {                   /* queue control block */
        struct  que_buf far    *qc_frst;    /* first on queue */
        struct  que_buf far    *qc_last;    /* last on queue */
} QC_;

typedef struct  rxd_data {      /* received data buffer */
        UTINY                                   rb_type;                /* message type */
        UTINY                   rb_line;        /* line number */
        time_t                  rb_time;        /* timestamp */
        UINT                    rb_stat;        /* message status flags */
        UINT                    rb_errs;        /* message error flags */
        UTINY                   rb_data[BUF_SIZE]; /* received data */
} RB_;

#define RB_HDR  (2*sizeof(UTINY) + 2*sizeof(UINT) + sizeof(time_t))

typedef struct line_ctl {       /* line control table */
        struct line_ctl *lc_next; /* next line control table in thread */
        struct device_ctl *lc_dvad; /* corresponding device control table */
        UINT          lc_errs;  /* receive error flags hold */
        int           lc_txcc;  /* Tx buffer count */
        UINT          lc_rxcc;  /* Rx buffer count */
        UINT          lc_trxm;  /* Rx total messages count */
        UINT          lc_obct;  /* out of buffers count */
        QC_           lc_txqc;  /* Tx queue control block */
        BF_    far   *lc_ctxb;  /* current Tx buffer */
        BF_    far   *lc_crxb;  /* current Rx buffer */
        UTINY  far   *lc_nxtc;  /* pointer to next Tx character */
        UTINY  far   *lc_nxrc;  /* pointer to next Rx character */
        time_t            lc_tfrx;      /* time first Rx char in buffer received */
        UTINY         lc_lnno;  /* line number */
        UTINY         lc_rxfl;  /* receive status flags hold */
        UTINY         lc_txst;  /* Tx state */
        UTINY         lc_rxst;  /* Rx state */
        UTINY         lc_txev;  /* Tx event signal */
        UTINY         lc_rxev;  /* Rx event signal */
} LC_;

typedef struct device_ctl {     /* device control table */
        struct device_ctl *dv_next; /* next device control table in thread */
        struct line_ctl *dv_lcad; /* corresponding line control table */
        UINT          dv_cmdp;  /* 8530 base port (commands) */
        UINT          dv_datp;  /* 8530 Tx/Rx data register */
        ULONG         dv_rbks;  /* receive break start, absolute ticks */
        ULONG         dv_tbks;  /* transmit break start */
        UTINY         dv_flgs;  /* flags */
        UTINY         dv_lrr0;  /* last RR0 status flags */
        UTINY         dv_txe;   /* transmit enable = 8, disable = 0 */
        UTINY         dv_rts;   /* RTS bit on = 2, off = 0 */
        UTINY         dv_dtr;   /* DTR bit on = 0x80, off = 0 */
        UTINY         dv_brks;  /* break status bit 0x10 = send bresk */

        /* configuration structure */

        UINT          dv_spd;   /* 16 line speed, bits/sec * 10 */
        UTINY         dv_bits;  /* 18 bits / character */
        UTINY         dv_par;   /* 19 parity: none (0), odd (1), even (3) */
        UTINY         dv_stpb;  /* 1A stop bits: 1 (4), 1.5 (8), 2 (12) */
        UTINY         dv_loop;  /* 1B local loopback 0 = no; 0x10 = yes */
} DV_;

typedef struct line_status_msg {        /* line status message */
        time_t              ls_time;        /* time of message */
        struct  device_ctl  ls_dctb;        /* port device control table */
        struct  line_ctl    ls_lctb;        /* port line control table */
} LS_;


                                        /*  ***** END OF FILE *****  */
