/**********************************************************************
*  File:      LC.C                                                    *
*  Purpose:   COMTROL Hostess i Programmer's Toolkit Main Line        *
*       Control Procedures                                            *
*  Date:      December 28, 1990                                       *
*  Author:    Howard D. Straus                                        *
*  Release:   1.00 - Original release                                 *
***********************************************************************
*                         COPYRIGHT NOTICE                            *
*=====================================================================*
*   Copyright 1990, 1991 by Howard D. Straus and Comtrol Corporation. *
*   All rights reserved.  Use or reproduction by any means of any     *
*   portion of this code contrary to the developers license agreement *
*   is expressly prohibited.                                          *
***********************************************************************/

#include "compiler.h"

#include <ctype.h>
#include <string.h>

#include "hostess.h"
#include "proto.h"
#include "externs.h"

void
lc_alc(lc)
        LC_             *lc;            /* line control table pointer */
{   int             i;              /* general loop variable */

        switch (lc->lc_rxst) {          /* action based on current Rx state */

        case CLOSED:                    /* line has been closed */
                lc->lc_rxst = lc_hcr(lc);   /* handle closed receiver */
                break;

        case IDLE:                      /* receiver open, unstarted */
                lc->lc_rxst = lc_rxs(lc);   /* attempt to start receive */
                break;

        case ACTIVE:                    /* receiver is active */
                lc->lc_rxst = lc_har(lc);   /* handle active receiver */
                break;

        case IN_BREAK:                  /* line in break */
                lc->lc_rxst = lc_wbe(lc);   /* wait for break to end */
                break;

        case LINE_FAIL:                 /* line failure */
                lc->lc_rxst = lc_hlf(lc);   /* handle line failure */
                break;

        default:                        /* unknown case */
                break;
        }

        switch (lc->lc_txst) {          /* see what xmtr needs to do */

        case CLOSED:                    /* transmitter closed */
                lc->lc_txst = lc_hct(lc);   /* handle closed transmitter */
                break;

        case IDLE:                      /* xmtr idle */
                lc->lc_txst = lc_hit(lc);   /* handle idle transmitter */
                break;

        case ACTIVE:                    /* transmitting */
                lc->lc_txst = lc_hat(lc);   /* handle active transmitter */
                break;

        case LINE_FAIL:                 /* line failure */
                lc->lc_txst = lc_hlf(lc);   /* handle line failure */
                break;

        default:
                break;
        }
}


/*
        close receiver
*/
int
lc_clr(lc)
        LC_             *lc;            /* line control table pointer */
{
        lc_rxf(lc);                     /* finish the current receive */
        dv_rxd(lc->lc_dvad);            /* disable the receiver */
        return(lc->lc_rxst = CLOSED);   /* change the Rx status to closed */
}


/*
        close transmitter
*/
int
lc_clt(lc)
        LC_             *lc;            /* line control table pointer */
{
        lc_txf(lc);                     /* finish the current transmission */
        dv_txd(lc->lc_dvad);            /* disable the transmitter */
        bq_fpq(&lc->lc_txqc);           /* free any buffers on the Tx queue */
        return(lc->lc_txst = CLOSED);   /* return closed status */
}


/*
        get next character to transmit
*/
int
lc_gnc(lc)
        LC_             *lc;            /* line control table address */
{   BF_     far     *bp;            /* temporary buffer hold */
        int             c;              /* character storage */

        c = *lc->lc_nxtc++;             /* get the next Tx character */
        if (--lc->lc_txcc <= 0)         /* is this buffer exhausted */
        {   bp = lc->lc_ctxb;           /* get pointer to local */
                bq_fre(bp);                 /* free the empty buffer */
                if ((bp = lc->lc_ctxb = bq_deq((QC_ far *)&lc->lc_txqc))
                        != NULL)                /* if another buffer to send */
                {   lc->lc_txcc = bp->bf_txct; /* set up char ct for next buffer */
                        lc->lc_nxtc = bp->bf_text; /* set next char pointer */
                }
                else                        /* no more buffers */
                        lc->lc_txcc = 0;        /* clear character count */
        }
        return(c);                      /* return the character */
}


/*
        handle active receiver
*/
int
lc_har(lc)
        LC_             *lc;
{       int                             r;                              /* return code */

        r = lc->lc_rxst;                                /* default return is current Rx status */
        switch (lc_wfr(lc)) {           /* check for receive events */

        case CL:                        /* host has closed line */
                lc_rxf(lc);                 /* terminate receive */
                dv_rxd(lc->lc_dvad);        /* disable receiver */
                ho_sls(lc);                 /* send host a line status message */
                r = CLOSED;                     /* return closed state */
                break;

        case TO:                        /* timeout occurred */
                lc->lc_errs |= ME_TIMEOUT;  /* timeout caused EOM */
                lc_rxf(lc);                 /* finish receive */
                r = lc_rxs(lc);                 /* arm for next receive */
                break;

        case BK:                        /* break started */
                ho_dig("Break start. //\r", lc->lc_lnno);
                lc->lc_errs |= ME_BRK_EOM;  /* break caused EOM */
                lc_rxf(lc);                 /* finish receive */
                ho_sls(lc);                 /* send host a line status message */
                r = IN_BREAK;                   /* put line in BREAK condition */
                break;

        case LE:                        /* line error */
                lc->lc_errs |= ME_BRK_EOM;  /* break caused EOM */
                ho_sls(lc);                 /* send host a line status message */
                r = LINE_FAIL;                  /* put line into failed state */
                break;

        case DC:                        /* DCD changed */
                lc->lc_errs |= ME_DCD_CHG;  /* put bit into message errors */
                r = lc_rxf(lc);             /* finish receive */
                ho_sle((lc->lc_dvad->dv_lrr0 & DCD ? LE_DCD_ACQUIRED : LE_DCD_LOST),
                                lc);                /* send notification of line change */
                ho_sls(lc);                 /* send line status */
                ho_dig((lc->lc_dvad->dv_lrr0 & DCD ? "Carrier Acquired. //\r" :
                                "Carrier Lost. //\r"), lc->lc_lnno);
                break;

        case CT:                        /* CTS changed */
                lc->lc_errs |= ME_CTS_CHG;  /* add to line errors */
                ho_sls(lc);                 /* send line status */
                ho_dig((lc->lc_dvad->dv_lrr0 & CTS ? "CTS Detected. //\r" :
                                "CTS Lost. //\r"), lc->lc_lnno);
                break;

        default:
                break;
        }

        return(r);                              /* return result of function */
}


/*
        handle active transmitter
*/
int                                 /* returns new Tx state */
lc_hat(lc)
        LC_             *lc;            /* line control table pointer */
{   int             r;              /* return state */

        r = lc->lc_txst;                /* get current line Tx state */

        switch (lc_wft(lc)) {           /* wait for transmit event */

        case TD:                        /* transmit done */
                r = IDLE;                   /* put line in idle state */
                break;

        case CL:                        /* host has closed line */
                r = lc->lc_txst = CLOSED;   /* fix state */
                dv_txd(lc->lc_dvad);        /* disable the transmitter */
                lc_txf(lc);                 /* release the buffers */
                dv_rts(0, lc->lc_dvad);     /* drop RTS */
                ho_sls(lc);                 /* send line status */
                break;

        case LE:                        /* line failed */
                r = LINE_FAIL;              /* return failure */
                ho_sls(lc);                 /* send line status */
                break;

        default:                        /* no changes */
                break;
        }
        return(r);                      /* return new state */
}


/*
        handle closed receiver
*/
int
lc_hcr(lc)
        LC_             *lc;
{
        switch (lc_wfr(lc)) {           /* wait for receive event */

        case OP:                        /* host opened receive */
                lc->lc_rxst = lc_rxs(lc);   /* start receive */
                ho_sls(lc);                 /* send line status */
                break;

        default:                        /* any other case */
                break;
        }
        return(lc->lc_rxst);
}


/*
        handle closed transmitter
*/
int
lc_hct(lc)
        LC_             *lc;
{   int             s;              /* save interrupt flag status */

        switch (lc_wft(lc)) {             /* wait for transmit event */

        case OP:                        /* host opened transmit */
                lc->lc_txst = IDLE;         /* move state to IDLE */
                ho_sls(lc);                 /* send line status */
                break;

        default:                        /* any other case */
                break;                      /* no change in state */
        }

        return(lc->lc_txst);
}


/*
        handle idle transmitter
*/
int
lc_hit(lc)
        LC_             *lc;
{   int             st;             /* current state */

        st = lc->lc_txst;               /* current state */
        switch (lc_wft(lc)) {           /* wait for transmit event */

        case CL:                        /* host has closed line */
                lc_txf(lc);                 /* terminate transmit */
                dv_rts(0, lc->lc_dvad);     /* drop RTS immediately */
                st = CLOSED;                /* change state */
                break;

        case ST:                        /* message exists, try to start */
                st = lc_txs(lc);            /* start transmission */
                break;

        default:
                break;
        }
        return(st);
}


/*
        handle line failure
*/
int
lc_hlf(lc)
        LC_             *lc;
{
        lc_rxf(lc);                       /* finish receive */
        ho_sle(LE_FAILURE, lc);           /* signal a line failure */
        ho_sls(lc);                       /* send line status */
        return(IDLE);                     /* return idle state */
}


/*
        finish receive
*/
int
lc_rxf(lc)
        LC_             *lc;
{   BF_     far     *bp;            /* buffer pointer */

        if ((bp = lc->lc_crxb) == NULL) /* current receive buffer */
                return(IDLE);               /* new line state */

        if (bp->bf_txct)                /* if any characters received */
        {   bp->bf_errs = lc->lc_errs;  /* transfer errors to buffer */
                bq_enq(lc->lc_crxb, (QC_ far *)&host_queue);
                                                                        /* queue message buffer to host */
        }
        else                            /* no characters received */
                bq_fre(bp);                 /* release empty buffer */
        lc->lc_crxb = NULL;             /* clear receive buffer pointer */
        return(lc->lc_rxst = IDLE);     /* put line in idle mode */
}


/*
        receive next character
*/
void
lc_rxn(c, lc)
        int             c;              /* character to analyze */
        LC_             *lc;
{   char            *ap;            /* for alignment function */
        int             i;              /* loop counter */

        lc_sch(c, lc);                  /* store the next character */
        if ((strchr("\r\n\03\04", c) != NULL) || /* if a special character */
                ((curr_time - lc->lc_crxb->bf_tmst) > 1)) /* or if buffer is old */
        {   lc_rxf(lc);                 /* send the line to the host */
                lc->lc_rxst = lc_rxs(lc);   /* start the next receive */
        }
}


/*
        start receive
*/
int
lc_rxs(lc)
        LC_             *lc;
{   BF_     far     *bp;            /* buffer pointer */

        if ((bp = bq_alc()) == NULL)    /* allocate a receive buffer */
        {   ho_alr(AL_NO_BUFFERS, lc);  /* send host alarm */
                return(CLOSED);             /* close line */
        }
        bp->bf_type = DATA_BUFFER;      /* indicate buffer type */
        bp->bf_lnno = lc->lc_lnno;      /* indicate line number */
        lc->lc_crxb = bp;               /* current Rx buffer */
        lc->lc_nxrc = bp->bf_text;      /* next receive character pointer */
        lc->lc_rxcc = BUF_SIZE;         /* size of buffer */
        lc->lc_tfrx = 0;                                /* clear buffer age indicator */
        dv_rxe(lc->lc_dvad);            /* enable receiver */
        return(ACTIVE);                 /* receiver waiting for input */
}


/*
        store character in receive buffer
*/
void
lc_sch(c, lc)
        int             c;              /* character to store */
        LC_             *lc;
{   BF_     far     *bp;            /* buffer pointer */

        *lc->lc_nxrc++ = c;             /* store in next available buffer slot */

        if (lc->lc_crxb->bf_txct++ == 0) /* if first character in buffer */
        {       lc->lc_crxb->bf_tmst = curr_time; /* timestamp it */
                lc->lc_tfrx = elapsed_ticks; /* remember buffer age */
        }

        if (--lc->lc_rxcc <= 0)         /* if buffer is full */
        {   if ((bp = bq_alc()) == NULL) /* allocate another */
                {   --lc->lc_nxrc;          /* drop character */
                        ++lc->lc_rxcc;          /* restore count */
                        ++lc->lc_obct;          /* increment out of buffers count */
                        return;
                }
                lc->lc_crxb->bf_errs = lc->lc_errs; /* save all errors */

                bp->bf_errs = lc->lc_errs;  /* transfer errors to new buffer */
                bq_enq(lc->lc_crxb, (QC_ far *)&host_queue); /* queue buff to host */

                lc->lc_crxb = bp;           /* make this current buffer */
                lc->lc_rxcc = BUF_SIZE;     /* initialize buffer space count */
                lc->lc_nxrc = bp->bf_text;  /* set next character pointer */
                bp->bf_type = DATA_BUFFER;  /* set buffer type */
                bp->bf_lnno = lc->lc_lnno;  /* add line number to buffer */
                bp->bf_txct = 0;            /* initialize count */
        }
}


/*
        signal receive event
*/
void
lc_sre(ev, lc)
        int             ev;             /* event to be signaled */
        LC_             *lc;
{
        lc->lc_rxev = ev;               /* store event */
}


/*
       signal transmit event
*/
void
lc_ste(ev, lc)
        int             ev;             /* event to be signaled */
        LC_             *lc;
{
        lc->lc_txev = ev;               /* store event */
}


/*
        finish transmission
*/
int
lc_txf(lc)
        LC_             *lc;
{
        bq_fre(lc->lc_ctxb);            /* release current Tx buffer */
        bq_fpq((QC_ far *)&lc->lc_txqc); /* release queued buffers */
        return(IDLE);                   /* return IDLE state */
}


/*
        start transmission
*/
int
lc_txs(lc)
        LC_             *lc;
{   BF_     far     *bp;            /* buffer pointer */
        int             st;             /* state to return */

        if ((bp = bq_deq((QC_ far *)&lc->lc_txqc)) == NULL) /* get data */
                return(IDLE);               /* no data buffer available */
        lc->lc_ctxb = bp;               /* current transmit buffer */
        lc->lc_txcc = bp->bf_txct;      /* number of chars in buffer */
        lc->lc_nxtc = bp->bf_text;      /* next transmit character */
        lc->lc_txst = ACTIVE;           /* set line active */
        dv_txs(lc->lc_dvad);            /* start transmitter */
        return(ACTIVE);                 /* return xmtr active */
}

/*
        wait for break to end
*/
int
lc_wbe(lc)
        LC_             *lc;
{   int                         r;                              /* result of function */

        r = lc->lc_rxst;                                /* default return is current status */
        switch (lc_wfr(lc)) {

        case BE:                        /* break ended */
                r = lc_rxs(lc);                 /* start receiver */
                ho_dig("Break End. //\r", lc->lc_lnno);
                break;

        default:                        /* anything else */
                break;
        }

        return(r);                                              /* return result of function */
}


/*
        wait for receive event
*/
int
lc_wfr(lc)
        LC_             *lc;
{   int     r;                      /* return value */

        if ((r = lc->lc_rxev) != 0)     /* event notification from elsewhere */
        {   lc->lc_rxev = 0;            /* clear event flag */
                return(r);                  /* return event */
        }

        if (lc->lc_rxst == ACTIVE)      /* if line is receiving */
        {   if (lc->lc_tfrx && (elapsed_ticks - lc->lc_tfrx) > 5)
                                                                        /* and buffer > 1/2 second old */
                        return(TO);             /* signal buffer age timeout */
        }

        return(NE);                     /* return no event detected */
}


/*
        wait for transmit event
*/
int
lc_wft(lc)
        LC_             *lc;
{   int     ev;                     /* event code */

        if ((ev = lc->lc_txev) != 0)    /* event has occurred */
        {   lc->lc_txev = 0;            /* clear the event */
                return(ev);                 /* return the event */
        }

        if (!bq_qem((QC_ far *)&lc->lc_txqc)) /* any message on the queue */
                return(ST);                 /* start the message */

        return(NE);                     /* return no event */
}


                                        /* ****** END OF FILE ***** */




